import bpy
from mathutils import Vector
from ..types import button
from ....icons import gizmos
from ....utils import addon


class NsuGizmo:
    @staticmethod
    def create(gizmo_group, context):
        props = addon.pref().tools.nsuite

        scale_factor = props.ops_gizmo_scale
        scale = 12 * scale_factor
        base_size = Vector((scale, scale))
        ui_scale = context.preferences.system.ui_scale
        base_space = 28 * scale_factor * ui_scale

        region = context.region
        area_width = region.width
        area_height = region.height

        # Calculate positions based on area size and DPI
        _location = Vector(props.ops_gizmo_location)
        xpos = (_location.x / 100.0) * area_width
        ypos = (_location.y / 100.0) * area_height

        location = Vector((xpos, ypos))
        flip = props.ops_gizmo_flip

        if flip:
            size_bar = Vector((scale / 4, scale))
        else:
            size_bar = Vector((scale, scale / 4))

        # Initial draw for the main button
        draw_button0 = button.Draw(gizmo_group, location, size_bar)
        draw_button0.operator("nsolve.transform_gizmo", properties={'gizmo': 'NSU'})

        analyze_running = False
        if hasattr(context.scene.nsolve, 'ops'):
            analyze_running = context.scene.nsolve.ops.mesh.analyze.running

        gizmo_definitions = {
            'hard_loop_cut': {
                'operator': "nsolve.hard_loop_cut",
                'properties': {'release_confirm': True},
                'icon': gizmos.hard_loop_cut.icon,
                'pushed': False,
            },
            'loop_cut': {
                'operator': "mesh.loopcut_slide",
                'properties': {},
                'icon': gizmos.loop_cut.icon,
                'pushed': False,
            },
            'hard_edge_slide': {
                'operator': "nsolve.hard_edge_slide",
                'properties': {},
                'icon': gizmos.hard_edge_slide.icon,
                'pushed': False,
            },
            'offset_edges': {
                'operator': "nsolve.offset_edges",
                'properties': {},
                'icon': gizmos.offset_edges.icon,
                'pushed': False,
            },
            'mesh_cut': {
                'operator': "nsolve.mesh_cut",
                'properties': {},
                'icon': gizmos.cut.icon,
                'pushed': False,
            },
            'bevel': {
                'operator': "nsolve.bevel",
                'properties': {},
                'icon': gizmos.bevel.icon,
                'pushed': False,
            },
            'project_line': {
                'operator': "nsolve.project_line",
                'properties': {'mouse_init': False},
                'icon': gizmos.project_line.icon,
                'pushed': False,
            },
            'project_loop': {
                'operator': "nsolve.project_loop",
                'properties': {'mode': 'OFFSET'},
                'icon': gizmos.project_loop.icon,
                'pushed': False,
            },
            'project_circle': {
                'operator': "nsolve.project_circle",
                'properties': {},
                'icon': gizmos.project_circle.icon,
                'pushed': False,
            },
            'project_grid': {
                'operator': "nsolve.project_grid",
                'properties': {},
                'icon': gizmos.project_grid.icon,
                'pushed': False,
            },
            'analyze': {
                'operator': "nsolve.analyze",
                'properties': {},
                'icon': gizmos.analyze.icon,
                'pushed': analyze_running,
            },
            'select_smartloop': {
                'operator': "nsolve.select_multi_smartloop",
                'properties': {},
                'icon': gizmos.select_smartloop.icon,
                'pushed': False,
            },
            'select_rounded_area': {
                'operator': "nsolve.select_rounded_area",
                'properties': {},
                'icon': gizmos.select_rounded_area.icon,
                'pushed': False,
            },
            'select_similar_strip': {
                'operator': "nsolve.select_similar_strip",
                'properties': {},
                'icon': gizmos.select_similar_strip.icon,
                'pushed': False,
            },
            'select_planarloop': {
                'operator': "nsolve.select_multi_planarloop",
                'properties': {},
                'icon': gizmos.select_planarloop.icon,
                'pushed': False,
            },
        }

        # Iterate over the current list items
        for item in props.ops_gizmo_list_items:
            gizmo_info = gizmo_definitions.get(item.operator)
            if gizmo_info:
                location = location.copy()
                if flip:
                    location.x += base_space
                else:
                    location.y -= base_space

                draw_button = button.Draw(gizmo_group, location, base_size, gizmo_info['icon'], gizmo_info['pushed'])
                draw_button.operator(gizmo_info['operator'], properties=gizmo_info['properties'])


class NSOLVE_GGT_Nsuite(bpy.types.GizmoGroup):
    bl_idname = 'NSOLVE_GGT_Nsuite'
    bl_label = 'Nsuite Gizmo'
    bl_space_type = 'VIEW_3D'
    bl_context_mode = 'EDIT_MESH'
    bl_region_type = 'WINDOW'
    bl_options = {'SHOW_MODAL_ALL', 'SCALE'}

    @classmethod
    def poll(cls, context):
        active_tool = bpy.context.workspace.tools.from_space_view3d_mode(bpy.context.mode, create=False)
        nsolve_tool = True if active_tool and active_tool.idname == 'nsolve.nsuitetool' else False
        return context.space_data.show_gizmo_tool and context.edit_object and nsolve_tool

    def setup(self, context):
        self.gizmos.clear()

    def refresh(self, context):
        self.gizmos.clear()

        nsuite = addon.pref().tools.nsuite
        if nsuite.nsu.hide:
            NsuGizmo.create(self, context)
