import bpy
from bpy_extras.io_utils import ExportHelper, ImportHelper
import json

from ....utils import addon


def snap_settings_to_dict(snap):
    return {
        "vert": snap.vert,
        "edge": snap.edge,
        "face": snap.face,
        "mid_edge": snap.mid_edge,
        "perp_edge": snap.perp_edge,
        "center_face": snap.center_face,
    }


def snap_settings_from_dict(snap, data):
    snap.vert = data.get("vert", True)
    snap.edge = data.get("edge", False)
    snap.face = data.get("face", False)
    snap.mid_edge = data.get("mid_edge", False)
    snap.perp_edge = data.get("perp_edge", False)
    snap.center_face = data.get("center_face", False)


def gizmo_collection_to_dict(collection):
    return {
        "operator": collection.operator,
        "select_mode": collection.select_mode,
        "multi_join": collection.multi_join,
        "snap_join": snap_settings_to_dict(collection.snap_join),
        "multi_merge": collection.multi_merge,
        "center_merge": collection.center_merge,
        "line_cut_through": collection.line_cut_through,
        "collapse_affect_selected": collection.collapse_affect_selected,
        "dissolve_use_verts": collection.dissolve_use_verts,
        "dissolve_use_face_split": collection.dissolve_use_face_split,
    }


def gizmo_collection_from_dict(collection, data):
    collection.operator = data.get("operator", 'JOIN')
    collection.select_mode = data.get("select_mode", 'SET')
    collection.multi_join = data.get("multi_join", False)
    snap_settings_from_dict(collection.snap_join, data.get("snap_join", {}))
    collection.multi_merge = data.get("multi_merge", False)
    collection.center_merge = data.get("center_merge", False)
    collection.line_cut_through = data.get("line_cut_through", False)
    collection.collapse_affect_selected = data.get("collapse_affect_selected", False)
    collection.dissolve_use_verts = data.get("dissolve_use_verts", True)
    collection.dissolve_use_face_split = data.get("dissolve_use_face_split", False)


class NSOLVE_OT_ExportPersona(bpy.types.Operator, ExportHelper):
    bl_idname = "nsolve.export_persona"
    bl_label = "Export Personas"
    bl_description = "Export the personas list to a .json file"

    filename_ext = ".nso"
    filter_glob: bpy.props.StringProperty(default="*.nso", options={'HIDDEN'})

    def execute(self, context):
        prop = addon.pref().tools.nsolve
        personas_list = [
            {
                "persona": item.persona,
                "main": gizmo_collection_to_dict(item.main),
                "aux": gizmo_collection_to_dict(item.aux),
                "sub": gizmo_collection_to_dict(item.sub),
                "ext": gizmo_collection_to_dict(item.ext),
            }
            for item in prop.personas_list_items
        ]

        data = {"personas_list_items": personas_list}

        with open(self.filepath, 'w') as f:
            json.dump(data, f, indent=4)

        self.report({'INFO'}, f"Exported {len(personas_list)} personas")
        return {'FINISHED'}


class NSOLVE_OT_ImportPersona(bpy.types.Operator, ImportHelper):
    bl_idname = "nsolve.import_persona"
    bl_label = "Import Personas"
    bl_description = "Import the personas list from a .json file"

    filename_ext = ".nso"
    filter_glob: bpy.props.StringProperty(default="*.nso", options={'HIDDEN'})

    def execute(self, context):
        prop = addon.pref().tools.nsolve

        with open(self.filepath, 'r') as f:
            data = json.load(f)

        personas_list = data.get("personas_list_items", [])

        prop.personas_list_items.clear()
        for item in personas_list:
            new_item = prop.personas_list_items.add()
            new_item.persona = item["persona"]
            gizmo_collection_from_dict(new_item.main, item["main"])
            gizmo_collection_from_dict(new_item.aux, item["aux"])
            gizmo_collection_from_dict(new_item.sub, item["sub"])
            gizmo_collection_from_dict(new_item.ext, item["ext"])

        self.report({'INFO'}, f"Imported {len(personas_list)} personas")
        return {'FINISHED'}
