import bpy
from mathutils import Vector
from ..utils import infobar
from ....utils import addon


class NSOLVE_OT_TransformGizmo(bpy.types.Operator):
    bl_idname = "nsolve.transform_gizmo"
    bl_label = "Move Gizmo"
    bl_description = "Move the gizmo in 3D space"
    bl_options = {'REGISTER', 'BLOCKING', 'GRAB_CURSOR', 'INTERNAL'}

    gizmo: bpy.props.EnumProperty(
        name="Gizmo",
        description="Gizmo",
        items=[
            ('NSO', 'nSuite', 'nSuite'),
            ('NSU', 'nSuite', 'nSuite')
        ],
        default='NSO'
    )

    mode: bpy.props.EnumProperty(
        name="Mode",
        description="Mode",
        items=[
            ('TRANSLATE', 'Translate', 'Translate the gizmo'),
            ('SCALE', 'Scale', 'Scale the gizmo')
        ],
        default='TRANSLATE'
    )

    start_mouse_pos: Vector = Vector()
    start_gizmo_pos: Vector = Vector()
    start_gizmo_scale: float = 1.0

    @classmethod
    def poll(cls, context):
        return context.area.type == 'VIEW_3D' and context.mode == 'EDIT_MESH'

    def invoke(self, context, event):
        self.start_mouse_pos = Vector((event.mouse_region_x, event.mouse_region_y))
        area_width = context.region.width
        area_height = context.region.height

        if self.gizmo == 'NSO':
            props = addon.pref().tools.nsolve
            self.start_gizmo_pos = Vector((
                (props.main_gizmo_location[0] / 100.0) * area_width,
                (props.main_gizmo_location[1] / 100.0) * area_height
            ))
            self.start_gizmo_scale = props.main_gizmo_scale
        else:
            props = addon.pref().tools.nsuite
            self.start_gizmo_pos = Vector((
                (props.ops_gizmo_location[0] / 100.0) * area_width,
                (props.ops_gizmo_location[1] / 100.0) * area_height
            ))
            self.start_gizmo_scale = props.ops_gizmo_scale

        infobar.draw(context, event, self.infobar_hotkeys, blank=True)

        self.mode = 'TRANSLATE'
        context.area.header_text_set('Transform Gizmo (Translate)')
        context.window.cursor_set('SCROLL_XY')

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def modal(self, context, event):
        if event.type == 'MOUSEMOVE':
            mouse_pos = Vector((event.mouse_region_x, event.mouse_region_y))
            if self.mode == 'TRANSLATE':
                self.set_gizmo_location(context, mouse_pos)
            elif self.mode == 'SCALE':
                self.set_gizmo_scale(context, mouse_pos)
            context.area.tag_redraw()

        elif event.type == 'LEFTMOUSE':
            self.end(context)
            return {'FINISHED'}

        elif event.type in {'RIGHTMOUSE', 'ESC'}:
            self.end(context)
            return {'CANCELLED'}

        elif event.type == 'G' and event.value == 'PRESS':
            self.mode = 'TRANSLATE'
            context.area.header_text_set('Transform Gizmo (Translate)')
        
        elif event.type == 'S' and event.value == 'PRESS':
            self.mode = 'SCALE'
            context.area.header_text_set('Transform Gizmo (Scale)')

        elif event.type == 'F' and event.value == 'PRESS':
            if self.gizmo == 'NSO':
                props = addon.pref().tools.nsolve
                props.main_gizmo_flip = not props.main_gizmo_flip
            else:
                props = addon.pref().tools.nsuite
                props.ops_gizmo_flip = not props.ops_gizmo_flip
            context.area.tag_redraw()

        return {'RUNNING_MODAL'}

    def set_gizmo_location(self, context, mouse_pos):
        delta = mouse_pos - self.start_mouse_pos
        location = self.start_gizmo_pos + delta

        width = context.region.width
        height = context.region.height

        # Clamp the location to within the region boundaries
        location[0] = max(25, min(width - 25, location[0]))
        location[1] = max(50, min(height - 25, location[1]))

        # Calculate percentage positions
        x_percentage = (location[0] / width) * 100
        y_percentage = (location[1] / height) * 100

        if self.gizmo == 'NSO':
            nsolve = addon.pref().tools.nsolve
            nsolve.main_gizmo_location = Vector((x_percentage, y_percentage))
        else:
            nsuite = addon.pref().tools.nsuite
            nsuite.ops_gizmo_location = Vector((x_percentage, y_percentage))

    def set_gizmo_scale(self, context, mouse_pos):
        delta = mouse_pos - self.start_mouse_pos
        scale_change = delta.x * 0.001  # Change this multiplier to adjust sensitivity
        new_scale = self.start_gizmo_scale + scale_change

        # Clamp the scale between 0.1 and 3
        new_scale = max(0.1, min(3, new_scale))

        if self.gizmo == 'NSO':
            addon.pref().tools.nsolve.main_gizmo_scale = new_scale
        else:
            addon.pref().tools.nsuite.ops_gizmo_scale = new_scale

    def end(self, context):
        '''Clean up.'''
        infobar.remove(context)
        context.area.header_text_set(text=None)
        context.window.cursor_set('CROSSHAIR')

    def infobar_hotkeys(self, layout, _context, _event):
        '''Draw the infobar with the hotkeys.'''
        row = layout.row(align=True)
        row.label(text='', icon='MOUSE_MOVE')
        row.label(text='Move / Scale Gizmo')
        row.separator(factor=12.0)
        row.label(text='', icon='MOUSE_LMB')
        row.label(text='Confirm')
        row.separator(factor=12.0)
        row.label(text='', icon='MOUSE_RMB')
        row.label(text='Cancel')
        row.separator(factor=12.0)
        row.label(text='G')
        row.label(text='Translate')
        row.separator(factor=12.0)
        row.label(text='S')
        row.label(text='Scale')


def register():
    bpy.utils.register_class(NSOLVE_OT_TransformGizmo)


def unregister():
    bpy.utils.unregister_class(NSOLVE_OT_TransformGizmo)
