import bpy
import bmesh
import blf
from ...draw.face import DrawFace
from ....utils import addon


class NSOLVE_OT_Analyze(bpy.types.Operator):
    bl_idname = "nsolve.analyze"
    bl_label = "Analyze"
    bl_description = f"{bl_label}\n\nDisplay visual information about the mesh"
    bl_options = {'REGISTER', 'UNDO'}

    offset: bpy.props.FloatProperty(name="Offset", description="Offset distance", default=0.1, min=0.0, precision=3)

    bm: bmesh.types.BMesh = None
    mesh: bpy.types.Mesh = None
    edges: list = []

    tris: list = []
    ngons: list = []
    ngon_tris: list = []
    tris_points: list = []

    _callback_ngon: DrawFace
    _handle_ngon: int

    _callback_tris: DrawFace
    _handle_tris: int

    _handle_ui: int

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

    @classmethod
    def poll(cls, context):
        return context.area.type == 'VIEW_3D' and context.mode == 'EDIT_MESH'

    def invoke(self, context, event):
        if context.scene.nsolve.ops.mesh.analyze.running:
            context.scene.nsolve.ops.mesh.analyze.running = False
            return {'CANCELLED'}

        context.scene.nsolve.ops.mesh.analyze.running = True

        obj = context.edit_object
        obj.update_from_editmode()

        matrix = obj.matrix_world

        self.update_mesh_info(obj.data, matrix)
        self.setup_drawing(context, self.ngon_tris, self.tris_points)

        bpy.app.handlers.depsgraph_update_post.append(self.depsgraph_update)

        context.window_manager.modal_handler_add(self)
        return {'RUNNING_MODAL'}

    def execute(self, context):
        return {'FINISHED'}

    def depsgraph_update(self, scene, depsgraph):
        obj = bpy.context.edit_object
        if not obj:
            return

        updated = False

        for update in depsgraph.updates:
            if isinstance(update.id, bpy.types.Object):
                updated_obj = update.id.original
                if updated_obj == obj:
                    updated = True

        if updated:
            obj.update_from_editmode()
            matrix = obj.matrix_world
            self.update_mesh_info(obj.data, matrix)
            self.update_drawing(bpy.context)

    def modal(self, context, event):
        '''Main modal loop.'''

        if not context.scene.nsolve.ops.mesh.analyze.running or not context.area:
            context.scene.nsolve.ops.mesh.analyze.running = False
            self.end(context)
            return {'CANCELLED'}

        # End the operator if we leave edit mode
        if context.mode != 'EDIT_MESH':
            context.scene.nsolve.ops.mesh.analyze.running = False
            self.end(context)
            return {'CANCELLED'}

        if event.type == 'MOUSEMOVE':
            context.area.tag_redraw()

        elif event.type in {'ESC'}:
            context.scene.nsolve.ops.mesh.analyze.running = False
            self.end(context)
            return {'CANCELLED'}

        return {'PASS_THROUGH'}

    def end(self, context):
        '''Clean up.'''

        self.edges = []

        self.ngon_tris = []
        self.tris_points = []
        self.ngons = []
        self.tris = []

        self.remove_draw_handlers()

        context.scene.nsolve.ops.mesh.analyze.running = False

        if self.depsgraph_update in bpy.app.handlers.depsgraph_update_post:
            bpy.app.handlers.depsgraph_update_post.remove(self.depsgraph_update)

        if context.area: context.area.tag_redraw()

    def remove_draw_handlers(self):
        bpy.types.SpaceView3D.draw_handler_remove(self._handle_ngon, 'WINDOW')
        bpy.types.SpaceView3D.draw_handler_remove(self._handle_tris, 'WINDOW')
        bpy.types.SpaceView3D.draw_handler_remove(self._handle_ui, 'WINDOW')

    def setup_drawing(self, context, ngons, tris):
        '''Setup the drawing'''

        theme = addon.pref().theme.ops.mesh.analyze

        self._handle_ui = bpy.types.SpaceView3D.draw_handler_add(self.draw_ui, (context,), 'WINDOW', 'POST_PIXEL')

        self._callback_ngon = DrawFace(self.ngon_tris, color=theme.ngon, culling=True)
        self._handle_ngon = bpy.types.SpaceView3D.draw_handler_add(self._callback_ngon.draw, (context,), 'WINDOW', 'POST_VIEW')

        self._callback_tris = DrawFace(self.tris_points, color=theme.tris, culling=True)
        self._handle_tris = bpy.types.SpaceView3D.draw_handler_add(self._callback_tris.draw, (context,), 'WINDOW', 'POST_VIEW')


    def update_drawing(self, context):
        '''Update the drawing with new points'''

        self._callback_ngon.update_batch(self.ngon_tris)
        self._callback_tris.update_batch(self.tris_points)

    def update_mesh_info(self, me, matrix):
        '''Update mesh info for drawing'''

        me.calc_loop_triangles()
        self.tris_points.clear()
        self.ngon_tris.clear()

        for i, loop in enumerate(me.loop_triangles):
            poly = me.polygons[me.loop_triangle_polygons[i].value]

            if poly.hide: continue

            if len(poly.vertices) == 3:
                tri = tuple(matrix @ me.vertices[j].co for j in loop.vertices)
                self.tris_points.append(tri)

            if len(poly.vertices) > 4:
                tri = tuple(matrix @ me.vertices[j].co for j in loop.vertices)
                self.ngon_tris.append(tri)

    def draw_ui(self, context):
        # Set position and size
        region = context.region
        width = region.width

        # x and y in percentage of width and height
        x = width - 170
        y = 70

        text = "Analyze"

        # Draw the text with GPU shaders
        font_id = 0
        blf.position(font_id, x, y, 0)
        blf.size(font_id, 20)
        blf.enable(font_id, blf.SHADOW)
        blf.shadow(font_id, 5, 0, 0, 0, 1)
        blf.color(font_id, 1, 1, 1, 0.6)
        blf.draw(font_id, text)


class types():
    class theme(bpy.types.PropertyGroup):
        ngon: bpy.props.FloatVectorProperty(name="Analyze Ngon", description="Color of Ngon", size=4, subtype='COLOR', default=(0.5, 0.0, 0.0, 0.1), min=0.0, max=1.0)
        tris: bpy.props.FloatVectorProperty(name="Analyze Tris", description="Color of Tris", size=4, subtype='COLOR', default=(0.5, 0.5, 0.0, 0.1), min=0.0, max=1.0)
        intersect: bpy.props.FloatVectorProperty(name="Analyze Intersect", description="Color of Intersect", size=4, subtype='COLOR', default=(0.0, 0.9, 0.0, 0.3), min=0.0, max=1.0)

    class scene(bpy.types.PropertyGroup):
        running: bpy.props.BoolProperty(name="Running", description="Operator is running", default=False)


types_classes = (types.theme,
                 types.scene,)
