from pathlib import Path
import bpy
from ...utils import addon
from .utils import refresh_gizmo
from ... import bl_info
from ...icons.menus import get_icon_id


def check_gizmo(context):
    area = context.area
    # Ensure the area is a 3D View
    if area.type == 'VIEW_3D':
        # Find the 'WINDOW' region within the 3D View
        region = next((r for r in area.regions if r.type == 'WINDOW'), None)
        if region:
            # Assuming 'nsolve' is set up correctly with ops_gizmo_location attributes
            nsuite = addon.pref().tools.nsuite
            locations = [nsuite.ops_gizmo_location]

            for location in locations:
                if location[0] < 5 or location[0] > 95 or location[1] < 0 or location[1] > 95:
                    return True

            return False

    return False  # If not a 3D View area or no 'WINDOW' region found


class NSOLVE_MT_Nsuite(bpy.types.WorkSpaceTool):
    bl_space_type = 'VIEW_3D'
    bl_context_mode = 'EDIT_MESH'
    bl_idname = 'nsolve.nsuitetool'
    bl_label = "nSuite"
    bl_description = f'Version {bl_info["version"][0]}.{bl_info["version"][1]}.{bl_info["version"][2]}'
    bl_icon = (Path(__file__).parent.parent.parent / "icons" / "tools" / "nsuite").as_posix()
    bl_widget = 'NSOLVE_GGT_Nsuite'
    bl_keymap = (
        ('view3d.select_box', {'type': 'LEFTMOUSE', 'value': 'CLICK_DRAG'}, {'properties': [('mode', 'SET')]}),
        ('view3d.select_box', {'type': 'LEFTMOUSE', 'value': 'CLICK_DRAG', 'ctrl': True}, {'properties': [('mode', 'SUB')]}),
        ('view3d.select_box', {'type': 'LEFTMOUSE', 'value': 'CLICK_DRAG', 'shift': True}, {'properties': [('mode', 'ADD')]}),
        ('nsolve.select_smartloop', {'type': 'RIGHTMOUSE', 'value': 'PRESS', 'ctrl': True}, {'properties': []}),
        ('nsolve.select_planarloop', {'type': 'RIGHTMOUSE', 'value': 'PRESS', 'shift': True}, {'properties': []}),
        ('nsolve.helper', {'type': 'D', 'value': 'PRESS'}, {'properties': []}),
    )

    def draw_settings(context, layout, tool):

        layout.popover("NSOLVE_PT_Nsuite", text="", icon='OPTIONS')
        row = layout.row(align=True)
        row.separator()
        row.popover("NSOLVE_PT_Options", text="", icon='PREFERENCES')

        out_of_bounds = check_gizmo(context)

        if out_of_bounds:
            row = layout.row(align=True)
            row.alert = True
            fix = row.operator("nsolve.fix_gizmo", text="Fix")
            fix.gizmo = 'NSU'

        if addon.pref().menus:
            layout.separator_spacer()

            if addon.pref().menus_type == 'SEPARATED':
                layout.menu("NSOLVE_MT_SelectMenu", text="Select")
                layout.menu("NSOLVE_MT_VertMenu", text="Vert")
                layout.menu("NSOLVE_MT_EdgeMenu", text="Edge")
                layout.menu("NSOLVE_MT_FaceMenu", text="Face")

            else:
                layout.menu("NSOLVE_MT_MainMenu", text="Menu")


def nsuite_options(layout):
    '''Draw the nsuite options'''

    nsuite = addon.pref().tools.nsuite

    layout.use_property_split = True
    layout.use_property_decorate = True

    col = layout

    box = col.column(align=True)
    subbox = box.box()
    col = subbox.column(align=True)
    row = col.row()
    row.template_list('NSOLVE_UL_NsuiteOpsGizmoList', '', nsuite, 'ops_gizmo_list_items', nsuite, 'ops_gizmo_list_index', rows=3)
    col = row.column(align=True)
    col.menu("NSOLVE_MT_add_operators_gizmo_menu", icon='ADD', text="")
    col.operator("nsolve.remove_operator", icon='REMOVE', text="")
    col.separator()
    col.operator("nsolve.move_operator_up", icon='TRIA_UP', text="")
    col.operator("nsolve.move_operator_down", icon='TRIA_DOWN', text="")
    col.separator()
    col.operator("nsolve.restore_default_operators", icon='EVENT_R', text="")


class NSOLVE_PT_Nsuite(bpy.types.Panel):
    bl_label = "nSolve"
    bl_idname = 'NSOLVE_PT_Nsuite'
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'WINDOW'

    def draw(self, context):

        layout = self.layout
        layout.ui_units_x = 14

        nsu = addon.pref().tools.nsuite.nsu
        maincol = layout.column(align=True)
        box = maincol.box()

        split = box.split(factor=0.7)
        row = split.row(align=True)
        row.use_property_split = False
        row.alignment = 'LEFT'
        row.prop(nsu, 'enable', text='', icon='TRIA_DOWN' if nsu.enable else 'TRIA_RIGHT', emboss=False)
        row.prop(nsu, 'enable', text='nSuite        ', emboss=False)

        row = split.row(align=True)
        row.alignment = 'RIGHT'
        row.operator("nsolve.place_gizmo", icon='DOT', text='').gizmo = 'NSU'
        row.separator()
        row.operator("nsolve.restore_default_operators", icon='EVENT_R', text="")
        row.separator()
        hide_icon = 'HIDE_OFF' if nsu.hide else 'HIDE_ON'
        row.prop(nsu, 'hide', text='', icon=hide_icon)

        if nsu.enable:
            nsuite_options(maincol)

        layout.separator(factor=2)
        split = layout.split(factor=0.9)
        info = f"{bl_info['version'][0]}.{bl_info['version'][1]}.{bl_info['version'][2]}"
        split.label(text=info)
        row = split.row(align=False)
        row.operator("nsolve.set_tool", text="", icon_value=get_icon_id("nSolve")).tool = 'NSO'


operators = [
    ('hard_loop_cut', 'Hard Loop Cut', 'Hard Loop Cut'),
    ('loop_cut', 'Loop Cut', 'Loop Cut'),
    ('hard_edge_slide', 'Hard Edge slide', 'Hard Edge slide'),
    ('offset_edges', 'Offset Edges', 'Offset Edges'),
    ('bevel', 'Bevel', 'Bevel'),
    ('mesh_cut', 'Mesh Cut', 'Mesh Cut'),
    ('project_line', 'Project Line', 'Project Line'),
    ('project_loop', 'Project Loop', 'Project Loop'),
    ('project_circle', 'Project Circle', 'Project Circle'),
    ('project_grid', 'Project Grid', 'Project Grid'),
    ('select_rounded_area', 'Select Rounded Area', 'Select Rounded Area'),
    ('select_similar_strip', 'Select Similar Strip', 'Select Similar Strip'),
    ('select_planarloop', 'Select Planar Loop', 'Select Planar Loop'),
    ('select_smartloop', 'Select Smart Loop', 'Select Smart Loop'),
    ('analyze', 'Analyze', 'Analyze'),]


class NSOLVE_OT_NsuiteMoveOperatorUp(bpy.types.Operator):
    bl_label = "Move Operator Up"
    bl_idname = "nsolve.move_operator_up"
    bl_description = ""
    bl_options = {'REGISTER', 'INTERNAL'}

    @classmethod
    def poll(cls, context):
        prop = addon.pref().tools.nsuite
        return prop.ops_gizmo_list_index > 0

    def execute(self, context):
        prop = addon.pref().tools.nsuite
        index = prop.ops_gizmo_list_index
        prop.ops_gizmo_list_items.move(index, index - 1)
        prop.ops_gizmo_list_index -= 1
        return {'FINISHED'}


class NSOLVE_OT_NsuiteMoveOperatorDown(bpy.types.Operator):
    bl_label = "Move Operator Down"
    bl_idname = "nsolve.move_operator_down"
    bl_options = {'REGISTER', 'INTERNAL'}
    bl_description = ""

    @classmethod
    def poll(cls, context):
        prop = addon.pref().tools.nsuite
        return prop.ops_gizmo_list_index < len(prop.ops_gizmo_list_items) - 1

    def execute(self, context):
        prop = addon.pref().tools.nsuite
        index = prop.ops_gizmo_list_index
        prop.ops_gizmo_list_items.move(index, index + 1)
        prop.ops_gizmo_list_index += 1
        return {'FINISHED'}


class NSOLVE_OT_NsuiteRemoveOperator(bpy.types.Operator):
    bl_label = "Remove Operator"
    bl_idname = "nsolve.remove_operator"
    bl_description = ""
    bl_options = {'REGISTER', 'INTERNAL'}

    @classmethod
    def poll(cls, context):
        prop = addon.pref().tools.nsuite
        return len(prop.ops_gizmo_list_items) > 0

    def execute(self, context):
        prop = addon.pref().tools.nsuite
        index = prop.ops_gizmo_list_index
        prop.ops_gizmo_list_items.remove(index)
        prop.ops_gizmo_list_index = min(max(0, index - 1), len(prop.ops_gizmo_list_items) - 1)
        return {'FINISHED'}


class NSOLVE_OT_NsuiteAddOperator(bpy.types.Operator):
    bl_label = "Add Operator"
    bl_idname = "nsolve.add_operator"
    bl_description = ""
    bl_options = {'REGISTER', 'INTERNAL'}

    operator: bpy.props.EnumProperty(name="Operator", description="Choose an operator to add", items=operators, default='hard_loop_cut')

    def execute(self, context):
        prop = addon.pref().tools.nsuite
        # Create a new gizmo item in the list
        new_item = prop.ops_gizmo_list_items.add()  # Adds a new item to the collection
        new_item.operator = self.operator  # Set the chosen operator to the new item
        prop.ops_gizmo_list_index = len(prop.ops_gizmo_list_items) - 1  # Optional: set the active index to the new item
        return {'FINISHED'}


class NSOLVE_OT_NsuiteRestoreDefaultOperartors(bpy.types.Operator):
    bl_label = "Restore Default Operators"
    bl_idname = "nsolve.restore_default_operators"
    bl_description = "Restore the default operators"
    bl_options = {'REGISTER', 'INTERNAL'}

    def execute(self, context):

        default_operators = (
            'hard_loop_cut',
            'loop_cut',
            'hard_edge_slide',
            'offset_edges',
            'bevel',
            'mesh_cut',
            'project_line',
            'project_loop',
            'project_circle',
            'project_grid',
            'select_smartloop',
            'analyze'
        )

        prop = addon.pref().tools.nsuite
        prop.ops_gizmo_list_items.clear()
        for operator in default_operators:
            new_item = prop.ops_gizmo_list_items.add()
            new_item.operator = operator

        prop.nsu.hide = True

        nsolve = addon.pref().tools.nsolve
        nsolve.nsu.hide = True

        return {'FINISHED'}


class NSOLVE_MT_NsuiteOperatorsAddMenu(bpy.types.Menu):
    bl_label = "Add Operator Menu"
    bl_idname = "NSOLVE_MT_add_operators_gizmo_menu"
    bl_description = ""

    def draw(self, context):
        layout = self.layout
        for ops_type, name, _ in operators:
            layout.operator("nsolve.add_operator", text=f"Add {name}").operator = ops_type


class NSOLVE_UL_NsuiteOpsGizmoList(bpy.types.UIList):
    def draw_item(self, context, layout, data, item, icon, active_data, active_propname, index):

        if item:
            op_name = next((name for code, name, desc in operators if code == item.operator), "Unknown")
            layout.label(text=f"{op_name}", translate=False)

    def draw_filter(self, context, layout):
        pass


class NsuiteOpsGizmoList(bpy.types.PropertyGroup):
    # Store the type of modifier
    operator: bpy.props.EnumProperty(name="Items", description="Items", items=operators, default='hard_loop_cut', update=refresh_gizmo)


class NsuiteNsuPanelPref(bpy.types.PropertyGroup):
    hide: bpy.props.BoolProperty(name="Hide", description="Hide", default=True, update=refresh_gizmo)
    enable: bpy.props.BoolProperty(name="Expand NSU", description="", default=False)
    layout: bpy.props.BoolProperty(name="Layout", description="Layout", default=True)


class types():
    class pref(bpy.types.PropertyGroup):
        ops_gizmo_location: bpy.props.FloatVectorProperty(name="Operators Gizmo Location", description="Gizmo Location", default=(45, 20), size=2, max=95, min=5, update=refresh_gizmo)
        ops_gizmo_scale: bpy.props.FloatProperty(name="Gizmo Scale", description="Gizmo Scale", default=1, min=0.5, max=3, update=refresh_gizmo)
        ops_gizmo_flip: bpy.props.BoolProperty(name="Flip", description="Flip", default=True, update=refresh_gizmo)
        ops_gizmo_list_items: bpy.props.CollectionProperty(type=NsuiteOpsGizmoList)
        ops_gizmo_list_index: bpy.props.IntProperty(name="Index", description="Index", default=-1, update=refresh_gizmo)

        nsu: bpy.props.PointerProperty(type=NsuiteNsuPanelPref)


types_classes = (
    NsuiteNsuPanelPref,
    NsuiteOpsGizmoList,
    NSOLVE_UL_NsuiteOpsGizmoList,
    types.pref,
)


classes = (
    NSOLVE_OT_NsuiteRestoreDefaultOperartors,
    NSOLVE_OT_NsuiteAddOperator,
    NSOLVE_OT_NsuiteMoveOperatorUp,
    NSOLVE_OT_NsuiteMoveOperatorDown,
    NSOLVE_OT_NsuiteRemoveOperator,
    NSOLVE_MT_NsuiteOperatorsAddMenu,
    NSOLVE_PT_Nsuite,
)


def initialize_nsuite_tool():

    default_operators = (
        'hard_loop_cut',
        'loop_cut',
        'hard_edge_slide',
        'offset_edges',
        'bevel',
        'mesh_cut',
        'project_line',
        'project_loop',
        'project_circle',
        'project_grid',
        'select_smartloop',
        'analyze'
    )

    prop = addon.pref().tools.nsuite
    if len(prop.ops_gizmo_list_items) == 0:
        for operator in default_operators:
            item = prop.ops_gizmo_list_items.add()
            item.operator = operator

    # Check if everything is set up as expected, then unregister the timer
    return None  # Returning None will stop the timer from repeating
